/* MD4 CUDA functions for cryptohaze.com Multiforcer */

typedef uint32_t UINT4;

/* MD4 Defines as per RFC reference implementation */
#define MD4F(x, y, z) (((x) & (y)) | ((~x) & (z)))
#define MD4G(x, y, z) (((x) & (y)) | ((x) & (z)) | ((y) & (z)))
#define MD4H(x, y, z) ((x) ^ (y) ^ (z))
#define MD4ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32-(n))))
#define MD4FF(a, b, c, d, x, s) { \
    (a) += MD4F ((b), (c), (d)) + (x); \
    (a) = MD4ROTATE_LEFT ((a), (s)); \
  }
#define MD4GG(a, b, c, d, x, s) { \
    (a) += MD4G ((b), (c), (d)) + (x) + (UINT4)0x5a827999; \
    (a) = MD4ROTATE_LEFT ((a), (s)); \
  }
#define MD4HH(a, b, c, d, x, s) { \
    (a) += MD4H ((b), (c), (d)) + (x) + (UINT4)0x6ed9eba1; \
    (a) = MD4ROTATE_LEFT ((a), (s)); \
  }
#define MD4S11 3
#define MD4S12 7
#define MD4S13 11
#define MD4S14 19
#define MD4S21 3
#define MD4S22 5
#define MD4S23 9
#define MD4S24 13
#define MD4S31 3
#define MD4S32 9
#define MD4S33 11
#define MD4S34 15
/* End MD4 Defines */

/*
 * MD4 code: Call as:
 * MD4(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, a, b, c, d); 
 * Takes inputs from b0-b15 and returns a, b, c, d.
*/
__device__ inline void CUDA_MD4(UINT4 b0, UINT4 b1, UINT4 b2, UINT4 b3, UINT4 b4, UINT4 b5, UINT4 b6, UINT4 b7, 
			   UINT4 b8, UINT4 b9, UINT4 b10, UINT4 b11, UINT4 b12, UINT4 b13, UINT4 b14, UINT4 b15, 
			   UINT4 &a, UINT4 &b, UINT4 &c, UINT4 &d) {
  a = 0x67452301;
  b = 0xefcdab89;
  c = 0x98badcfe;
  d = 0x10325476;

  MD4FF (a, b, c, d, b0, MD4S11); /* 1 */
  MD4FF (d, a, b, c, b1, MD4S12); /* 2 */
  MD4FF (c, d, a, b, b2, MD4S13); /* 3 */
  MD4FF (b, c, d, a, b3, MD4S14); /* 4 */
  MD4FF (a, b, c, d, b4, MD4S11); /* 5 */
  MD4FF (d, a, b, c, b5, MD4S12); /* 6 */
  MD4FF (c, d, a, b, b6, MD4S13); /* 7 */
  MD4FF (b, c, d, a, b7, MD4S14); /* 8 */
  MD4FF (a, b, c, d, b8, MD4S11); /* 9 */
  MD4FF (d, a, b, c, b9, MD4S12); /* 10 */
  MD4FF (c, d, a, b, b10, MD4S13); /* 11 */
  MD4FF (b, c, d, a, b11, MD4S14); /* 12 */
  MD4FF (a, b, c, d, b12, MD4S11); /* 13 */
  MD4FF (d, a, b, c, b13, MD4S12); /* 14 */
  MD4FF (c, d, a, b, b14, MD4S13); /* 15 */
  MD4FF (b, c, d, a, b15, MD4S14); /* 16 */

  /* Round 2 */
  MD4GG (a, b, c, d, b0, MD4S21); /* 17 */
  MD4GG (d, a, b, c, b4, MD4S22); /* 18 */
  MD4GG (c, d, a, b, b8, MD4S23); /* 19 */
  MD4GG (b, c, d, a, b12, MD4S24); /* 20 */
  MD4GG (a, b, c, d, b1, MD4S21); /* 21 */
  MD4GG (d, a, b, c, b5, MD4S22); /* 22 */
  MD4GG (c, d, a, b, b9, MD4S23); /* 23 */
  MD4GG (b, c, d, a, b13, MD4S24); /* 24 */
  MD4GG (a, b, c, d, b2, MD4S21); /* 25 */
  MD4GG (d, a, b, c, b6, MD4S22); /* 26 */
  MD4GG (c, d, a, b, b10, MD4S23); /* 27 */
  MD4GG (b, c, d, a, b14, MD4S24); /* 28 */
  MD4GG (a, b, c, d, b3, MD4S21); /* 29 */
  MD4GG (d, a, b, c, b7, MD4S22); /* 30 */
  MD4GG (c, d, a, b, b11, MD4S23); /* 31 */
  MD4GG (b, c, d, a, b15, MD4S24); /* 32 */


  /* Round 3 */
  MD4HH (a, b, c, d, b0, MD4S31); /* 33 */
  MD4HH (d, a, b, c, b8, MD4S32); /* 34 */
  MD4HH (c, d, a, b, b4, MD4S33); /* 35 */
  MD4HH (b, c, d, a, b12, MD4S34); /* 36 */
  MD4HH (a, b, c, d, b2, MD4S31); /* 37 */
  MD4HH (d, a, b, c, b10, MD4S32); /* 38 */
  MD4HH (c, d, a, b, b6, MD4S33); /* 39 */
  MD4HH (b, c, d, a, b14, MD4S34); /* 40 */
  MD4HH (a, b, c, d, b1, MD4S31); /* 41 */
  MD4HH (d, a, b, c, b9, MD4S32); /* 42 */
  MD4HH (c, d, a, b, b5, MD4S33); /* 43 */
  MD4HH (b, c, d, a, b13, MD4S34); /* 44 */
  MD4HH (a, b, c, d, b3, MD4S31); /* 45 */
  MD4HH (d, a, b, c, b11, MD4S32); /* 46 */
  MD4HH (c, d, a, b, b7, MD4S33); /* 47 */
  MD4HH (b, c, d, a, b15, MD4S34); /* 48 */

  // Finally, add initial values, as this is the only pass we make.
  a += 0x67452301;
  b += 0xefcdab89;
  c += 0x98badcfe;
  d += 0x10325476;  
}



void CPU_MD4(unsigned char *toHash, unsigned int length, unsigned char *hashResult) {
  // 32-bit unsigned values for the hash
  UINT4 a,b,c,d;
  UINT4 b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15;
  
  UINT4 *InitialArray32;
  UINT4 *OutputArray32;
  // 32-bit accesses to the hash arrays
  InitialArray32 = (UINT4 *)toHash;
  OutputArray32 = (UINT4 *)hashResult;


  
  b0 = (UINT4)InitialArray32[0];
  b1 = (UINT4)InitialArray32[1];
  b2 = (UINT4)InitialArray32[2];
  b3 = (UINT4)InitialArray32[3];
  b4 = (UINT4)InitialArray32[4];
  b5 = (UINT4)InitialArray32[5];
  b6 = (UINT4)InitialArray32[6];
  b7 = (UINT4)InitialArray32[7];
  b8 = (UINT4)InitialArray32[8];
  b9 = (UINT4)InitialArray32[9];
  b10 = (UINT4)InitialArray32[10];
  b11 = (UINT4)InitialArray32[11];
  b12 = (UINT4)InitialArray32[12];
  b13 = (UINT4)InitialArray32[13];
  b14 = (UINT4)InitialArray32[14];
  
  switch (length) {
    case 0:
	  b0 |= 0x00000080;
	  break;
	case 1:
	  b0 |= 0x00008000;
	  break;
	case 2:
	  b0 |= 0x00800000;
	  break;
	case 3:
	  b0 |= 0x80000000;
	  break;
    case 4:
	  b1 |= 0x00000080;
	  break;
	case 5:
	  b1 |= 0x00008000;
	  break;
	case 6:
	  b1 |= 0x00800000;
	  break;
	case 7:
	  b1 |= 0x80000000;
	  break;
    case 8:
	  b2 |= 0x00000080;
	  break;
	case 9:
	  b2 |= 0x00008000;
	  break;
	case 10:
	  b2 |= 0x00800000;
	  break;
	case 11:
	  b2 |= 0x80000000;
	  break;
    case 12:
	  b3 |= 0x00000080;
	  break;
	case 13:
	  b3 |= 0x00008000;
	  break;
	case 14:
	  b3 |= 0x00800000;
	  break;
	case 15:
	  b3 |= 0x80000000;
	  break;
    case 16:
	  b4 |= 0x00000080;
	  break;
	case 17:
	  b4 |= 0x00008000;
	  break;
	case 18:
	  b4 |= 0x00800000;
	  break;
	case 19:
	  b4 |= 0x80000000;
	  break;

  }
  b14 = length * 8;
  b15 = 0x00000000;
  
  a = 0x67452301;
  b = 0xefcdab89;
  c = 0x98badcfe;
  d = 0x10325476;

  MD4FF (a, b, c, d, b0, MD4S11); /* 1 */
  MD4FF (d, a, b, c, b1, MD4S12); /* 2 */
  MD4FF (c, d, a, b, b2, MD4S13); /* 3 */
  MD4FF (b, c, d, a, b3, MD4S14); /* 4 */
  MD4FF (a, b, c, d, b4, MD4S11); /* 5 */
  MD4FF (d, a, b, c, b5, MD4S12); /* 6 */
  MD4FF (c, d, a, b, b6, MD4S13); /* 7 */
  MD4FF (b, c, d, a, b7, MD4S14); /* 8 */
  MD4FF (a, b, c, d, b8, MD4S11); /* 9 */
  MD4FF (d, a, b, c, b9, MD4S12); /* 10 */
  MD4FF (c, d, a, b, b10, MD4S13); /* 11 */
  MD4FF (b, c, d, a, b11, MD4S14); /* 12 */
  MD4FF (a, b, c, d, b12, MD4S11); /* 13 */
  MD4FF (d, a, b, c, b13, MD4S12); /* 14 */
  MD4FF (c, d, a, b, b14, MD4S13); /* 15 */
  MD4FF (b, c, d, a, b15, MD4S14); /* 16 */

  /* Round 2 */
  MD4GG (a, b, c, d, b0, MD4S21); /* 17 */
  MD4GG (d, a, b, c, b4, MD4S22); /* 18 */
  MD4GG (c, d, a, b, b8, MD4S23); /* 19 */
  MD4GG (b, c, d, a, b12, MD4S24); /* 20 */
  MD4GG (a, b, c, d, b1, MD4S21); /* 21 */
  MD4GG (d, a, b, c, b5, MD4S22); /* 22 */
  MD4GG (c, d, a, b, b9, MD4S23); /* 23 */
  MD4GG (b, c, d, a, b13, MD4S24); /* 24 */
  MD4GG (a, b, c, d, b2, MD4S21); /* 25 */
  MD4GG (d, a, b, c, b6, MD4S22); /* 26 */
  MD4GG (c, d, a, b, b10, MD4S23); /* 27 */
  MD4GG (b, c, d, a, b14, MD4S24); /* 28 */
  MD4GG (a, b, c, d, b3, MD4S21); /* 29 */
  MD4GG (d, a, b, c, b7, MD4S22); /* 30 */
  MD4GG (c, d, a, b, b11, MD4S23); /* 31 */
  MD4GG (b, c, d, a, b15, MD4S24); /* 32 */


  /* Round 3 */
  MD4HH (a, b, c, d, b0, MD4S31); /* 33 */
  MD4HH (d, a, b, c, b8, MD4S32); /* 34 */
  MD4HH (c, d, a, b, b4, MD4S33); /* 35 */
  MD4HH (b, c, d, a, b12, MD4S34); /* 36 */
  MD4HH (a, b, c, d, b2, MD4S31); /* 37 */
  MD4HH (d, a, b, c, b10, MD4S32); /* 38 */
  MD4HH (c, d, a, b, b6, MD4S33); /* 39 */
  MD4HH (b, c, d, a, b14, MD4S34); /* 40 */
  MD4HH (a, b, c, d, b1, MD4S31); /* 41 */
  MD4HH (d, a, b, c, b9, MD4S32); /* 42 */
  MD4HH (c, d, a, b, b5, MD4S33); /* 43 */
  MD4HH (b, c, d, a, b13, MD4S34); /* 44 */
  MD4HH (a, b, c, d, b3, MD4S31); /* 45 */
  MD4HH (d, a, b, c, b11, MD4S32); /* 46 */
  MD4HH (c, d, a, b, b7, MD4S33); /* 47 */
  MD4HH (b, c, d, a, b15, MD4S34); /* 48 */

  // Finally, add initial values, as this is the only pass we make.
  a += 0x67452301;
  b += 0xefcdab89;
  c += 0x98badcfe;
  d += 0x10325476;
    
  
  OutputArray32[0] = a;
  OutputArray32[1] = b;
  OutputArray32[2] = c;
  OutputArray32[3] = d;

}





int CUDA_MD4_Get_Thread_Count(int passlength) {
  int threads;
  switch (passlength) {
	case 4:
	case 5: 
	case 6: 
	  threads = 512;
	  break;
	case 7: 
	case 8: 
	  threads = 448;
	  break;
	case 9: 
  	case 10: 
	case 11: 
	  threads = 384;
	  break;
	case 12: 
	case 13: 
	case 14: 
	  threads = 320;
	  break;
  }
  return threads;
}





void CPU_MD4_Search_0(unsigned char *SearchHash, unsigned char *Output, unsigned char *success, unsigned char *global_charset, int charsetLen, int numberOfPasswords) {

  unsigned char hash[16];
  unsigned char passwordString[64];
  int password;
  memset(passwordString, 0, 64);
  CPU_MD4(passwordString, 0, hash);
    // Password created.  Check against arrays.
    for (password = 0; password < numberOfPasswords; password++) {
      // If there is a match, deal with it appropriately.
      if (memcmp(hash, &SearchHash[HASH_LENGTH*password], HASH_LENGTH) == 0) {
        success[password] = 1;
      }
    }

}
/*

__global__ void CUDA_MD4_Search_1(unsigned char *OutputPassword, unsigned char *success, 
			    int charsetLen, uint32_t numberOfPasswords, struct start_positions *DEVICE_Start_Positions, unsigned int count, unsigned char * DEVICE_Hashes, unsigned char *DEVICE_HashTable) {
  
  const int pass_length = 1;
  
  // Hash work space: 16 by 32-bit registers
  uint32_t b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15;
  // Output space: 4 by 32-bit registers
  uint32_t a,b,c,d;
  // Thread index, for start position loads
  uint32_t thread_index = blockIdx.x*blockDim.x + threadIdx.x;
  
  uint32_t *DEVICE_Hashes_32 = (uint32_t *)DEVICE_Hashes;
  
  // Password charset counters
  unsigned char p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15;
  UINT4 password_count = 0;
  __shared__ __align__(16) unsigned char sharedCharset[4096];
  __shared__ __align__(16) unsigned char sharedBitmap[8192];
  __shared__ unsigned char sharedLengths[16];
  
  copyCharsetAndBitmap(sharedCharset, sharedBitmap, sharedLengths, charsetLen, pass_length);

  // Load start positions from global memory into device registers
  loadStartPositions(pass_length, thread_index, DEVICE_Start_Positions, 
		   p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15);
  
  while (password_count < count) {

  //initMD(pass_length, (unsigned char *)deviceCharset,
  initMD(pass_length, sharedCharset,
  	p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
	b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15);	
#ifdef __DEVICE_EMULATION__
  printf("Thread %d: Indexes: %d%d%d%d%d%d\n",thread_index, p0, p1, p2, p3, p4, p5);
  printf("Password: %c%c%c%c%c%c\n", sharedCharset[p0],  sharedCharset[p1 + 256],  sharedCharset[p2 + 512],  sharedCharset[p3 + 768],  sharedCharset[p4 + 1024],  sharedCharset[p5 + 1280]);
  printf("%08X %08X\n", b0, b1);
#endif
  // Hash function, inlined.
  CUDA_MD4(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, a, b, c, d);

  // Check for valid results.
  
  checkHashMulti(pass_length, sharedBitmap, DEVICE_HashTable, numberOfPasswords,
		DEVICE_Hashes_32, success, OutputPassword,
		p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
		a, b, c, d, b0, b1, b2, b3, b4, b5);
  password_count++;
  
  incrementCounters1Multi();
  }
}

__global__ void CUDA_MD4_Search_2(unsigned char *OutputPassword, unsigned char *success, 
			    int charsetLen, uint32_t numberOfPasswords, struct start_positions *DEVICE_Start_Positions, unsigned int count, unsigned char * DEVICE_Hashes, unsigned char *DEVICE_HashTable) {
  
  const int pass_length = 2;
  
  // Hash work space: 16 by 32-bit registers
  uint32_t b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15;
  // Output space: 4 by 32-bit registers
  uint32_t a,b,c,d;
  // Thread index, for start position loads
  uint32_t thread_index = blockIdx.x*blockDim.x + threadIdx.x;
  
  uint32_t *DEVICE_Hashes_32 = (uint32_t *)DEVICE_Hashes;
  
  // Password charset counters
  unsigned char p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15;
  UINT4 password_count = 0;
  __shared__ __align__(16) unsigned char sharedCharset[4096];
  __shared__ __align__(16) unsigned char sharedBitmap[8192];
  __shared__ unsigned char sharedLengths[16];
  
  copyCharsetAndBitmap(sharedCharset, sharedBitmap, sharedLengths, charsetLen, pass_length);

  // Load start positions from global memory into device registers
  loadStartPositions(pass_length, thread_index, DEVICE_Start_Positions, 
		   p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15);
  
  while (password_count < count) {

  //initMD(pass_length, (unsigned char *)deviceCharset,
  initMD(pass_length, sharedCharset,
  	p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
	b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15);	
#ifdef __DEVICE_EMULATION__
  printf("Thread %d: Indexes: %d%d%d%d%d%d\n",thread_index, p0, p1, p2, p3, p4, p5);
  printf("Password: %c%c%c%c%c%c\n", sharedCharset[p0],  sharedCharset[p1 + 256],  sharedCharset[p2 + 512],  sharedCharset[p3 + 768],  sharedCharset[p4 + 1024],  sharedCharset[p5 + 1280]);
  printf("%08X %08X\n", b0, b1);
#endif
  // Hash function, inlined.
  CUDA_MD4(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, a, b, c, d);

  // Check for valid results.
  
  checkHashMulti(pass_length, sharedBitmap, DEVICE_HashTable, numberOfPasswords,
		DEVICE_Hashes_32, success, OutputPassword,
		p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
		a, b, c, d, b0, b1, b2, b3, b4, b5);
  password_count++;
  
  incrementCounters2Multi();
  }
}

__global__ void CUDA_MD4_Search_3(unsigned char *OutputPassword, unsigned char *success, 
			    int charsetLen, uint32_t numberOfPasswords, struct start_positions *DEVICE_Start_Positions, unsigned int count, unsigned char * DEVICE_Hashes, unsigned char *DEVICE_HashTable) {
  
  const int pass_length = 3;
  
  // Hash work space: 16 by 32-bit registers
  uint32_t b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15;
  // Output space: 4 by 32-bit registers
  uint32_t a,b,c,d;
  // Thread index, for start position loads
  uint32_t thread_index = blockIdx.x*blockDim.x + threadIdx.x;
  
  uint32_t *DEVICE_Hashes_32 = (uint32_t *)DEVICE_Hashes;
  
  // Password charset counters
  unsigned char p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15;
  UINT4 password_count = 0;
  __shared__ __align__(16) unsigned char sharedCharset[4096];
  __shared__ __align__(16) unsigned char sharedBitmap[8192];
  __shared__ unsigned char sharedLengths[16];
  
  copyCharsetAndBitmap(sharedCharset, sharedBitmap, sharedLengths, charsetLen, pass_length);

  // Load start positions from global memory into device registers
  loadStartPositions(pass_length, thread_index, DEVICE_Start_Positions, 
		   p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15);
  
  while (password_count < count) {

  //initMD(pass_length, (unsigned char *)deviceCharset,
  initMD(pass_length, sharedCharset,
  	p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
	b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15);	
#ifdef __DEVICE_EMULATION__
  printf("Thread %d: Indexes: %d%d%d%d%d%d\n",thread_index, p0, p1, p2, p3, p4, p5);
  printf("Password: %c%c%c%c%c%c\n", sharedCharset[p0],  sharedCharset[p1 + 256],  sharedCharset[p2 + 512],  sharedCharset[p3 + 768],  sharedCharset[p4 + 1024],  sharedCharset[p5 + 1280]);
  printf("%08X %08X\n", b0, b1);
#endif
  // Hash function, inlined.
  CUDA_MD4(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, a, b, c, d);

  // Check for valid results.
  
  checkHashMulti(pass_length, sharedBitmap, DEVICE_HashTable, numberOfPasswords,
		DEVICE_Hashes_32, success, OutputPassword,
		p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
		a, b, c, d, b0, b1, b2, b3, b4, b5);
  password_count++;
  
  incrementCounters3Multi();
  }
}

__global__ void CUDA_MD4_Search_4(unsigned char *OutputPassword, unsigned char *success, 
			    int charsetLen, uint32_t numberOfPasswords, struct start_positions *DEVICE_Start_Positions, unsigned int count, unsigned char * DEVICE_Hashes, unsigned char *DEVICE_HashTable) {
  
  const int pass_length = 4;
  
  // Hash work space: 16 by 32-bit registers
  uint32_t b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15;
  // Output space: 4 by 32-bit registers
  uint32_t a,b,c,d;
  // Thread index, for start position loads
  uint32_t thread_index = blockIdx.x*blockDim.x + threadIdx.x;
  
  uint32_t *DEVICE_Hashes_32 = (uint32_t *)DEVICE_Hashes;
  
  // Password charset counters
  unsigned char p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15;
  UINT4 password_count = 0;
  __shared__ __align__(16) unsigned char sharedCharset[4096];
  __shared__ __align__(16) unsigned char sharedBitmap[8192];
  __shared__ unsigned char sharedLengths[16];
  
  copyCharsetAndBitmap(sharedCharset, sharedBitmap, sharedLengths, charsetLen, pass_length);

  // Load start positions from global memory into device registers
  loadStartPositions(pass_length, thread_index, DEVICE_Start_Positions, 
		   p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15);
  
  while (password_count < count) {

  //initMD(pass_length, (unsigned char *)deviceCharset,
  initMD(pass_length, sharedCharset,
  	p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
	b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15);	
#ifdef __DEVICE_EMULATION__
  printf("Thread %d: Indexes: %d%d%d%d%d%d\n",thread_index, p0, p1, p2, p3, p4, p5);
  printf("Password: %c%c%c%c%c%c\n", sharedCharset[p0],  sharedCharset[p1 + 256],  sharedCharset[p2 + 512],  sharedCharset[p3 + 768],  sharedCharset[p4 + 1024],  sharedCharset[p5 + 1280]);
  printf("%08X %08X\n", b0, b1);
#endif
  // Hash function, inlined.
  CUDA_MD4(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, a, b, c, d);

  // Check for valid results.
  
  checkHashMulti(pass_length, sharedBitmap, DEVICE_HashTable, numberOfPasswords,
		DEVICE_Hashes_32, success, OutputPassword,
		p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
		a, b, c, d, b0, b1, b2, b3, b4, b5);
  password_count++;
  
  incrementCounters4Multi();
  }
}



__global__ void CUDA_MD4_Search_5(unsigned char *OutputPassword, unsigned char *success, 
			    int charsetLen, uint32_t numberOfPasswords, struct start_positions *DEVICE_Start_Positions, unsigned int count, unsigned char * DEVICE_Hashes, unsigned char *DEVICE_HashTable) {
  
  const int pass_length = 5;
  
  // Hash work space: 16 by 32-bit registers
  uint32_t b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15;
  // Output space: 4 by 32-bit registers
  uint32_t a,b,c,d;
  // Thread index, for start position loads
  uint32_t thread_index = blockIdx.x*blockDim.x + threadIdx.x;
  
  uint32_t *DEVICE_Hashes_32 = (uint32_t *)DEVICE_Hashes;
  
  // Password charset counters
  unsigned char p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15;
  UINT4 password_count = 0;
  __shared__ __align__(16) unsigned char sharedCharset[4096];
  __shared__ __align__(16) unsigned char sharedBitmap[8192];
  __shared__ unsigned char sharedLengths[16];
  
  copyCharsetAndBitmap(sharedCharset, sharedBitmap, sharedLengths, charsetLen, pass_length);

  // Load start positions from global memory into device registers
  loadStartPositions(pass_length, thread_index, DEVICE_Start_Positions, 
		   p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15);
  
  while (password_count < count) {

  //initMD(pass_length, (unsigned char *)deviceCharset,
  initMD(pass_length, sharedCharset,
  	p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
	b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15);	
#ifdef __DEVICE_EMULATION__
  printf("Thread %d: Indexes: %d%d%d%d%d%d\n",thread_index, p0, p1, p2, p3, p4, p5);
  printf("Password: %c%c%c%c%c%c\n", sharedCharset[p0],  sharedCharset[p1 + 256],  sharedCharset[p2 + 512],  sharedCharset[p3 + 768],  sharedCharset[p4 + 1024],  sharedCharset[p5 + 1280]);
  printf("%08X %08X\n", b0, b1);
#endif
  // Hash function, inlined.
  CUDA_MD4(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, a, b, c, d);

  // Check for valid results.
  
  checkHashMulti(pass_length, sharedBitmap, DEVICE_HashTable, numberOfPasswords,
		DEVICE_Hashes_32, success, OutputPassword,
		p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
		a, b, c, d, b0, b1, b2, b3, b4, b5);
  password_count++;
  
  incrementCounters5Multi();
  }
}


__global__ void CUDA_MD4_Search_6(unsigned char *OutputPassword, unsigned char *success, 
			    int charsetLen, uint32_t numberOfPasswords, struct start_positions *DEVICE_Start_Positions, unsigned int count, unsigned char * DEVICE_Hashes, unsigned char *DEVICE_HashTable) {
  
  const int pass_length = 6;
  
  // Hash work space: 16 by 32-bit registers
  uint32_t b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15;
  // Output space: 4 by 32-bit registers
  uint32_t a,b,c,d;
  // Thread index, for start position loads
  uint32_t thread_index = blockIdx.x*blockDim.x + threadIdx.x;
  
  uint32_t *DEVICE_Hashes_32 = (uint32_t *)DEVICE_Hashes;
  
  // Password charset counters
  unsigned char p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15;
  UINT4 password_count = 0;
  __shared__ __align__(16) unsigned char sharedCharset[4096];
  __shared__ __align__(16) unsigned char sharedBitmap[8192];
  __shared__ unsigned char sharedLengths[16];
  
  copyCharsetAndBitmap(sharedCharset, sharedBitmap, sharedLengths, charsetLen, pass_length);

  // Load start positions from global memory into device registers
  loadStartPositions(pass_length, thread_index, DEVICE_Start_Positions, 
		   p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15);
  
  while (password_count < count) {

  //initMD(pass_length, (unsigned char *)deviceCharset,
  initMD(pass_length, sharedCharset,
  	p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
	b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15);	
#ifdef __DEVICE_EMULATION__
  printf("Thread %d: Indexes: %d%d%d%d%d%d\n",thread_index, p0, p1, p2, p3, p4, p5);
  printf("Password: %c%c%c%c%c%c\n", sharedCharset[p0],  sharedCharset[p1 + 256],  sharedCharset[p2 + 512],  sharedCharset[p3 + 768],  sharedCharset[p4 + 1024],  sharedCharset[p5 + 1280]);
  printf("%08X %08X\n", b0, b1);
#endif
  // Hash function, inlined.
  CUDA_MD4(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, a, b, c, d);

  // Check for valid results.
  
  checkHashMulti(pass_length, sharedBitmap, DEVICE_HashTable, numberOfPasswords,
		DEVICE_Hashes_32, success, OutputPassword,
		p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
		a, b, c, d, b0, b1, b2, b3, b4, b5);
  password_count++;
  
  incrementCounters6Multi();
  }
}


__global__ void CUDA_MD4_Search_7(unsigned char *OutputPassword, unsigned char *success, 
			    int charsetLen, uint32_t numberOfPasswords, struct start_positions *DEVICE_Start_Positions, unsigned int count, unsigned char * DEVICE_Hashes, unsigned char *DEVICE_HashTable) {
  
  
  const int pass_length = 7;
  
  // Hash work space: 16 by 32-bit registers
  uint32_t b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15;
  // Output space: 4 by 32-bit registers
  uint32_t a,b,c,d;
  // Thread index, for start position loads
  uint32_t thread_index = blockIdx.x*blockDim.x + threadIdx.x;
  
  uint32_t *DEVICE_Hashes_32 = (uint32_t *)DEVICE_Hashes;
  
  // Password charset counters
  unsigned char p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15;
  UINT4 password_count = 0;
  __shared__ __align__(16) unsigned char sharedCharset[4096];
  __shared__ __align__(16) unsigned char sharedBitmap[8192];
  __shared__ __align__(8) unsigned char sharedLengths[16];
  
  copyCharsetAndBitmap(sharedCharset, sharedBitmap, sharedLengths, charsetLen, pass_length);

  // Load start positions from global memory into device registers
  loadStartPositions(pass_length, thread_index, DEVICE_Start_Positions, 
		   p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15);
  
  while (password_count < count) {

  //initMD(pass_length, (unsigned char *)deviceCharset,
  initMD(pass_length, sharedCharset,
  	p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
	b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15);	

  // Hash function, inlined.
  CUDA_MD4(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, a, b, c, d);

  // Check for valid results.
  
  checkHashMulti(pass_length, sharedBitmap, DEVICE_HashTable, numberOfPasswords,
		DEVICE_Hashes_32, success, OutputPassword,
		p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
		a, b, c, d, b0, b1, b2, b3, b4, b5);
  password_count++;
  
  incrementCounters7Multi();
  }
}



__global__ void CUDA_MD4_Search_8(unsigned char *OutputPassword, unsigned char *success, 
			    int charsetLen, uint32_t numberOfPasswords, struct start_positions *DEVICE_Start_Positions, unsigned int count, unsigned char * DEVICE_Hashes, unsigned char *DEVICE_HashTable) {
  
  
  const int pass_length = 8;
  
  // Hash work space: 16 by 32-bit registers
  uint32_t b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15;
  // Output space: 4 by 32-bit registers
  uint32_t a,b,c,d;
  // Thread index, for start position loads
  uint32_t thread_index = blockIdx.x*blockDim.x + threadIdx.x;
  
  uint32_t *DEVICE_Hashes_32 = (uint32_t *)DEVICE_Hashes;
  
  // Password charset counters
  unsigned char p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15;
  UINT4 password_count = 0;
  __shared__ __align__(16) unsigned char sharedCharset[4096];
  __shared__ __align__(16) unsigned char sharedBitmap[8192];
  __shared__ __align__(8) unsigned char sharedLengths[16];
  
  copyCharsetAndBitmap(sharedCharset, sharedBitmap, sharedLengths, charsetLen, pass_length);

  // Load start positions from global memory into device registers
  loadStartPositions(pass_length, thread_index, DEVICE_Start_Positions, 
		   p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15);
  
  while (password_count < count) {

  //initMD(pass_length, (unsigned char *)deviceCharset,
  initMD(pass_length, sharedCharset,
  	p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
	b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15);	

  // Hash function, inlined.
  CUDA_MD4(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, a, b, c, d);

  // Check for valid results.
  
  checkHashMulti(pass_length, sharedBitmap, DEVICE_HashTable, numberOfPasswords,
		DEVICE_Hashes_32, success, OutputPassword,
		p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
		a, b, c, d, b0, b1, b2, b3, b4, b5);
  password_count++;
  
  incrementCounters8Multi();
  }
}


__global__ void CUDA_MD4_Search_9(unsigned char *OutputPassword, unsigned char *success, 
			    int charsetLen, uint32_t numberOfPasswords, struct start_positions *DEVICE_Start_Positions, unsigned int count, unsigned char * DEVICE_Hashes, unsigned char *DEVICE_HashTable) {
  const int pass_length = 9;
  uint32_t b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15;
  uint32_t a,b,c,d;
  uint32_t thread_index = blockIdx.x*blockDim.x + threadIdx.x;
  uint32_t *DEVICE_Hashes_32 = (uint32_t *)DEVICE_Hashes;
  unsigned char p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15;
  UINT4 password_count = 0;
  __shared__ __align__(16) unsigned char sharedCharset[4096];
  __shared__ __align__(16) unsigned char sharedBitmap[8192];
  __shared__ __align__(8) unsigned char sharedLengths[16];
  copyCharsetAndBitmap(sharedCharset, sharedBitmap, sharedLengths, charsetLen, pass_length);
  loadStartPositions(pass_length, thread_index, DEVICE_Start_Positions, 
		   p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15);
  while (password_count < count) {
  initMD(pass_length, sharedCharset,
  	p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
	b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15);	
  CUDA_MD4(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, a, b, c, d);
  checkHashMulti(pass_length, sharedBitmap, DEVICE_HashTable, numberOfPasswords,
		DEVICE_Hashes_32, success, OutputPassword,
		p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
		a, b, c, d, b0, b1, b2, b3, b4, b5);
  password_count++;
  incrementCounters9Multi();
  }
}

__global__ void CUDA_MD4_Search_10(unsigned char *OutputPassword, unsigned char *success, 
			    int charsetLen, uint32_t numberOfPasswords, struct start_positions *DEVICE_Start_Positions, unsigned int count, unsigned char * DEVICE_Hashes, unsigned char *DEVICE_HashTable) {
  const int pass_length = 10;
  uint32_t b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15;
  uint32_t a,b,c,d;
  uint32_t thread_index = blockIdx.x*blockDim.x + threadIdx.x;
  uint32_t *DEVICE_Hashes_32 = (uint32_t *)DEVICE_Hashes;
  unsigned char p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15;
  UINT4 password_count = 0;
  __shared__ __align__(16) unsigned char sharedCharset[4096];
  __shared__ __align__(16) unsigned char sharedBitmap[8192];
  __shared__ __align__(8) unsigned char sharedLengths[16];
  copyCharsetAndBitmap(sharedCharset, sharedBitmap, sharedLengths, charsetLen, pass_length);
  loadStartPositions(pass_length, thread_index, DEVICE_Start_Positions, 
		   p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15);
  while (password_count < count) {
  initMD(pass_length, sharedCharset,
  	p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
	b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15);	
  CUDA_MD4(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, a, b, c, d);
  checkHashMulti(pass_length, sharedBitmap, DEVICE_HashTable, numberOfPasswords,
		DEVICE_Hashes_32, success, OutputPassword,
		p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
		a, b, c, d, b0, b1, b2, b3, b4, b5);
  password_count++;
  incrementCounters10Multi();
  }
}

__global__ void CUDA_MD4_Search_11(unsigned char *OutputPassword, unsigned char *success, 
			    int charsetLen, uint32_t numberOfPasswords, struct start_positions *DEVICE_Start_Positions, unsigned int count, unsigned char * DEVICE_Hashes, unsigned char *DEVICE_HashTable) {
  const int pass_length = 11;
  uint32_t b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15;
  uint32_t a,b,c,d;
  uint32_t thread_index = blockIdx.x*blockDim.x + threadIdx.x;
  uint32_t *DEVICE_Hashes_32 = (uint32_t *)DEVICE_Hashes;
  unsigned char p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15;
  UINT4 password_count = 0;
  __shared__ __align__(16) unsigned char sharedCharset[4096];
  __shared__ __align__(16) unsigned char sharedBitmap[8192];
  __shared__ __align__(8) unsigned char sharedLengths[16];
  copyCharsetAndBitmap(sharedCharset, sharedBitmap, sharedLengths, charsetLen, pass_length);
  loadStartPositions(pass_length, thread_index, DEVICE_Start_Positions, 
		   p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15);
  while (password_count < count) {
  initMD(pass_length, sharedCharset,
  	p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
	b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15);	
  CUDA_MD4(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, a, b, c, d);
  checkHashMulti(pass_length, sharedBitmap, DEVICE_HashTable, numberOfPasswords,
		DEVICE_Hashes_32, success, OutputPassword,
		p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
		a, b, c, d, b0, b1, b2, b3, b4, b5);
  password_count++;
  incrementCounters11Multi();
  }
}

__global__ void CUDA_MD4_Search_12(unsigned char *OutputPassword, unsigned char *success, 
			    int charsetLen, uint32_t numberOfPasswords, struct start_positions *DEVICE_Start_Positions, unsigned int count, unsigned char * DEVICE_Hashes, unsigned char *DEVICE_HashTable) {
  const int pass_length = 12;
  uint32_t b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15;
  uint32_t a,b,c,d;
  uint32_t thread_index = blockIdx.x*blockDim.x + threadIdx.x;
  uint32_t *DEVICE_Hashes_32 = (uint32_t *)DEVICE_Hashes;
  unsigned char p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15;
  UINT4 password_count = 0;
  __shared__ __align__(16) unsigned char sharedCharset[4096];
  __shared__ __align__(16) unsigned char sharedBitmap[8192];
  __shared__ __align__(8) unsigned char sharedLengths[16];
  copyCharsetAndBitmap(sharedCharset, sharedBitmap, sharedLengths, charsetLen, pass_length);
  loadStartPositions(pass_length, thread_index, DEVICE_Start_Positions, 
		   p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15);
  while (password_count < count) {
  initMD(pass_length, sharedCharset,
  	p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
	b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15);	
  CUDA_MD4(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, a, b, c, d);
  checkHashMulti(pass_length, sharedBitmap, DEVICE_HashTable, numberOfPasswords,
		DEVICE_Hashes_32, success, OutputPassword,
		p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
		a, b, c, d, b0, b1, b2, b3, b4, b5);
  password_count++;
  incrementCounters12Multi();
  }
}
__global__ void CUDA_MD4_Search_13(unsigned char *OutputPassword, unsigned char *success, 
			    int charsetLen, uint32_t numberOfPasswords, struct start_positions *DEVICE_Start_Positions, unsigned int count, unsigned char * DEVICE_Hashes, unsigned char *DEVICE_HashTable) {
  const int pass_length = 13;
  uint32_t b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15;
  uint32_t a,b,c,d;
  uint32_t thread_index = blockIdx.x*blockDim.x + threadIdx.x;
  uint32_t *DEVICE_Hashes_32 = (uint32_t *)DEVICE_Hashes;
  unsigned char p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15;
  UINT4 password_count = 0;
  __shared__ __align__(16) unsigned char sharedCharset[4096];
  __shared__ __align__(16) unsigned char sharedBitmap[8192];
  __shared__ __align__(8) unsigned char sharedLengths[16];
  copyCharsetAndBitmap(sharedCharset, sharedBitmap, sharedLengths, charsetLen, pass_length);
  loadStartPositions(pass_length, thread_index, DEVICE_Start_Positions, 
		   p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15);
  while (password_count < count) {
  initMD(pass_length, sharedCharset,
  	p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
	b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15);	
  CUDA_MD4(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, a, b, c, d);
  checkHashMulti(pass_length, sharedBitmap, DEVICE_HashTable, numberOfPasswords,
		DEVICE_Hashes_32, success, OutputPassword,
		p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
		a, b, c, d, b0, b1, b2, b3, b4, b5);
  password_count++;
  incrementCounters13Multi();
  }
}
__global__ void CUDA_MD4_Search_14(unsigned char *OutputPassword, unsigned char *success, 
			    int charsetLen, uint32_t numberOfPasswords, struct start_positions *DEVICE_Start_Positions, unsigned int count, unsigned char * DEVICE_Hashes, unsigned char *DEVICE_HashTable) {
  const int pass_length = 14;
  uint32_t b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15;
  uint32_t a,b,c,d;
  uint32_t thread_index = blockIdx.x*blockDim.x + threadIdx.x;
  uint32_t *DEVICE_Hashes_32 = (uint32_t *)DEVICE_Hashes;
  unsigned char p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15;
  UINT4 password_count = 0;
  __shared__ __align__(16) unsigned char sharedCharset[4096];
  __shared__ __align__(16) unsigned char sharedBitmap[8192];
  __shared__ __align__(8) unsigned char sharedLengths[16];
  copyCharsetAndBitmap(sharedCharset, sharedBitmap, sharedLengths, charsetLen, pass_length);
  loadStartPositions(pass_length, thread_index, DEVICE_Start_Positions, 
		   p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15);
  while (password_count < count) {
  initMD(pass_length, sharedCharset,
  	p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
	b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15);	
  CUDA_MD4(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, a, b, c, d);
  checkHashMulti(pass_length, sharedBitmap, DEVICE_HashTable, numberOfPasswords,
		DEVICE_Hashes_32, success, OutputPassword,
		p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
		a, b, c, d, b0, b1, b2, b3, b4, b5);
  password_count++;
  incrementCounters14Multi();
  }
}
__global__ void CUDA_MD4_Search_15(unsigned char *OutputPassword, unsigned char *success, 
			    int charsetLen, uint32_t numberOfPasswords, struct start_positions *DEVICE_Start_Positions, unsigned int count, unsigned char * DEVICE_Hashes, unsigned char *DEVICE_HashTable) {
  const int pass_length = 15;
  uint32_t b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15;
  uint32_t a,b,c,d;
  uint32_t thread_index = blockIdx.x*blockDim.x + threadIdx.x;
  uint32_t *DEVICE_Hashes_32 = (uint32_t *)DEVICE_Hashes;
  unsigned char p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15;
  UINT4 password_count = 0;
  __shared__ __align__(16) unsigned char sharedCharset[4096];
  __shared__ __align__(16) unsigned char sharedBitmap[8192];
  __shared__ __align__(8) unsigned char sharedLengths[16];
  copyCharsetAndBitmap(sharedCharset, sharedBitmap, sharedLengths, charsetLen, pass_length);
  loadStartPositions(pass_length, thread_index, DEVICE_Start_Positions, 
		   p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15);
  while (password_count < count) {
  initMD(pass_length, sharedCharset,
  	p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
	b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15);	
  CUDA_MD4(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, a, b, c, d);
  checkHashMulti(pass_length, sharedBitmap, DEVICE_HashTable, numberOfPasswords,
		DEVICE_Hashes_32, success, OutputPassword,
		p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
		a, b, c, d, b0, b1, b2, b3, b4, b5);
  password_count++;
  incrementCounters15Multi();
  }
}
__global__ void CUDA_MD4_Search_16(unsigned char *OutputPassword, unsigned char *success, 
			    int charsetLen, uint32_t numberOfPasswords, struct start_positions *DEVICE_Start_Positions, unsigned int count, unsigned char * DEVICE_Hashes, unsigned char *DEVICE_HashTable) {
  const int pass_length = 16;
  uint32_t b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15;
  uint32_t a,b,c,d;
  uint32_t thread_index = blockIdx.x*blockDim.x + threadIdx.x;
  uint32_t *DEVICE_Hashes_32 = (uint32_t *)DEVICE_Hashes;
  unsigned char p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15;
  UINT4 password_count = 0;
  __shared__ __align__(16) unsigned char sharedCharset[4096];
  __shared__ __align__(16) unsigned char sharedBitmap[8192];
  __shared__ __align__(8) unsigned char sharedLengths[16];
  copyCharsetAndBitmap(sharedCharset, sharedBitmap, sharedLengths, charsetLen, pass_length);
  loadStartPositions(pass_length, thread_index, DEVICE_Start_Positions, 
		   p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15);
  while (password_count < count) {
  initMD(pass_length, sharedCharset,
  	p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
	b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15);	
  CUDA_MD4(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, a, b, c, d);
  checkHashMulti(pass_length, sharedBitmap, DEVICE_HashTable, numberOfPasswords,
		DEVICE_Hashes_32, success, OutputPassword,
		p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15,
		a, b, c, d, b0, b1, b2, b3, b4, b5);
  password_count++;
  incrementCounters16Multi();
  }
}

*/

#define MD4_CUDA_KERNEL_CREATE(length) \
__global__ void CUDA_MD4_Search_##length (unsigned char *OutputPassword, unsigned char *success,  \
			    int charsetLen, uint32_t numberOfPasswords, struct start_positions *DEVICE_Start_Positions, \
				unsigned int count, unsigned char * DEVICE_Hashes, unsigned char *DEVICE_HashTable) { \
  const int pass_length = length; \
  uint32_t b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15; \
  uint32_t a,b,c,d; \
  uint32_t thread_index = blockIdx.x*blockDim.x + threadIdx.x; \
  uint32_t *DEVICE_Hashes_32 = (uint32_t *)DEVICE_Hashes; \
  unsigned char p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15; \
  UINT4 password_count = 0; \
  __shared__ __align__(16) unsigned char sharedCharset[4096]; \
  __shared__ __align__(16) unsigned char sharedBitmap[8192]; \
  __shared__ unsigned char sharedLengths[16]; \
  copyCharsetAndBitmap(sharedCharset, sharedBitmap, sharedLengths, charsetLen, pass_length); \
  loadStartPositions(pass_length, thread_index, DEVICE_Start_Positions,  \
		   p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15); \
  while (password_count < count) { \
  initMD(pass_length, sharedCharset, \
  	p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15, \
	b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15);	 \
  CUDA_MD4(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, a, b, c, d); \
  checkHashMulti(pass_length, sharedBitmap, DEVICE_HashTable, numberOfPasswords, \
		DEVICE_Hashes_32, success, OutputPassword, \
		p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15, \
		a, b, c, d, b0, b1, b2, b3, b4, b5); \
  password_count++; \
  incrementCounters##length##Multi(); \
  } \
}

#define FASTMD4_CUDA_KERNEL_CREATE(length) \
__global__ void CUDA_FASTMD4_Search_##length (unsigned char *OutputPassword, unsigned char *success,  \
			    int charsetLen, uint32_t numberOfPasswords, struct start_positions *DEVICE_Start_Positions,  \
				unsigned int count, unsigned char * DEVICE_Hashes, unsigned char *DEVICE_HashTable) { \
  const int pass_length = length; \
  uint32_t b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15; \
  uint32_t a,b,c,d; \
  uint32_t thread_index = blockIdx.x*blockDim.x + threadIdx.x; \
  uint32_t *DEVICE_Hashes_32 = (uint32_t *)deviceSearchHashes; \
  unsigned char p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15; \
  UINT4 password_count = 0; \
  __shared__ __align__(16) unsigned char sharedCharset[256]; \
  copyCharsetFAST(sharedCharset, charsetLen); \
  loadStartPositions(pass_length, thread_index, DEVICE_Start_Positions,  \
		   p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15); \
  while (password_count < count) { \
  initMDFAST(pass_length, sharedCharset, \
  	p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15, \
	b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15);	 \
  CUDA_MD4(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, a, b, c, d); \
  checkHashFAST(pass_length, numberOfPasswords, \
		DEVICE_Hashes_32, success, OutputPassword, \
		p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15, \
		a, b, c, d, b0); \
  password_count++; \
  incrementCounters##length (); \
  } \
}

MD4_CUDA_KERNEL_CREATE(1)
MD4_CUDA_KERNEL_CREATE(2)
MD4_CUDA_KERNEL_CREATE(3)
MD4_CUDA_KERNEL_CREATE(4)
MD4_CUDA_KERNEL_CREATE(5)
MD4_CUDA_KERNEL_CREATE(6)
MD4_CUDA_KERNEL_CREATE(7)
MD4_CUDA_KERNEL_CREATE(8)
MD4_CUDA_KERNEL_CREATE(9)
MD4_CUDA_KERNEL_CREATE(10)
MD4_CUDA_KERNEL_CREATE(11)
MD4_CUDA_KERNEL_CREATE(12)
MD4_CUDA_KERNEL_CREATE(13)
MD4_CUDA_KERNEL_CREATE(14)
MD4_CUDA_KERNEL_CREATE(15)
MD4_CUDA_KERNEL_CREATE(16)

FASTMD4_CUDA_KERNEL_CREATE(1)
FASTMD4_CUDA_KERNEL_CREATE(2)
FASTMD4_CUDA_KERNEL_CREATE(3)
FASTMD4_CUDA_KERNEL_CREATE(4)
FASTMD4_CUDA_KERNEL_CREATE(5)
FASTMD4_CUDA_KERNEL_CREATE(6)
FASTMD4_CUDA_KERNEL_CREATE(7)
FASTMD4_CUDA_KERNEL_CREATE(8)
FASTMD4_CUDA_KERNEL_CREATE(9)
FASTMD4_CUDA_KERNEL_CREATE(10)
FASTMD4_CUDA_KERNEL_CREATE(11)
FASTMD4_CUDA_KERNEL_CREATE(12)
FASTMD4_CUDA_KERNEL_CREATE(13)
FASTMD4_CUDA_KERNEL_CREATE(14)
FASTMD4_CUDA_KERNEL_CREATE(15)
FASTMD4_CUDA_KERNEL_CREATE(16)

void Launch_CPU_MD4_Kernel(int passlength, unsigned char *SearchHash, unsigned char *Output, unsigned char *success, 
							unsigned char *global_charset, int charsetLen, int numberOfPasswords) {
  if (passlength == 0) {
    CPU_MD4_Search_0(SearchHash, Output, success, global_charset, charsetLen, numberOfPasswords);
  }
}


void Launch_CUDA_MD4_Kernel(int passlength, uint32_t charsetLength, int numberOfPasswords, unsigned char *DEVICE_Passwords, 
						unsigned char *DEVICE_Success, struct start_positions *DEVICE_Start_Positions, uint32_t per_step, uint32_t threads, uint32_t blocks, unsigned char *DEVICE_Hashes, unsigned char *DEVICE_Bitmap) {
    if (passlength == 1) {
	  CUDA_MD4_Search_1 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 2) {
	  CUDA_MD4_Search_2 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 3) {
	  CUDA_MD4_Search_3 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 4) {
	  CUDA_MD4_Search_4 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 5) {
	  CUDA_MD4_Search_5 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 6) {
	  CUDA_MD4_Search_6 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 7) {
	  CUDA_MD4_Search_7 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 8) {
	  CUDA_MD4_Search_8 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);  
	} else if (passlength == 9) {
	  CUDA_MD4_Search_9 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 10) {
	  CUDA_MD4_Search_10 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 11) {
	  CUDA_MD4_Search_11 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 12) {
	  CUDA_MD4_Search_12 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 13) {
	  CUDA_MD4_Search_13 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 14) {
	  CUDA_MD4_Search_14 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 15) {
	  CUDA_MD4_Search_15 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 16) {
	  CUDA_MD4_Search_16 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	}
	
	cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err) 
      {
        fprintf(stderr, "Cuda error: %s.\n", cudaGetErrorString( err) );
        exit(EXIT_FAILURE);
      } 
}


void Launch_CUDA_FASTMD4_Kernel(int passlength, uint32_t charsetLength, int numberOfPasswords, unsigned char *DEVICE_Passwords, 
						unsigned char *DEVICE_Success, struct start_positions *DEVICE_Start_Positions, uint32_t per_step, uint32_t threads, uint32_t blocks, unsigned char *DEVICE_Hashes, unsigned char *DEVICE_Bitmap) {
    if (passlength == 1) {
	  CUDA_MD4_Search_1 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 2) {
	  CUDA_MD4_Search_2 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 3) {
	  CUDA_MD4_Search_3 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 4) {
	  CUDA_MD4_Search_4 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 5) {
	  CUDA_MD4_Search_5 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 6) {
	  CUDA_MD4_Search_6 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 7) {
	  CUDA_MD4_Search_7 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 8) {
	  CUDA_MD4_Search_8 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);  
	} else if (passlength == 9) {
	  CUDA_MD4_Search_9 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 10) {
	  CUDA_MD4_Search_10 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 11) {
	  CUDA_MD4_Search_11 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 12) {
	  CUDA_MD4_Search_12 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 13) {
	  CUDA_MD4_Search_13 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 14) {
	  CUDA_MD4_Search_14 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 15) {
	  CUDA_MD4_Search_15 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 16) {
	  CUDA_MD4_Search_16 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	}
	
	cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err) 
      {
        fprintf(stderr, "Cuda error: %s.\n", cudaGetErrorString( err) );
        exit(EXIT_FAILURE);
      } 
}

