#include "./argtable2-9/src/argtable2.h"
#include <cuda.h>
#include <cutil.h>
#include "Multiforcer.h"


// Max number of hashes to support - default 4000 (Slightly less than the max 64kb of constant memory)
#define MAX_HASHES 2000
// Charset size - right now Unicode is not supported, so 256 characters covers everything
#define MAX_CHARSET 256
// Hash length in bytes - this code supports MD4/MD5/NTLM so 128 bit hash output
#define HASH_LENGTH 16

// Max length we support
// Not correct, but leave it for now, as the NTLM, MD4, and MD5 brute forcers are hard coded for
// 16 right now - changing this breaks their reporting.
#define MAX_PASSWORD_LENGTH 16

// GLobal char for verbose
unsigned char verbose = NULL;
unsigned int big_lookup_table = 0;


// Device constant memory allocation for hashes & charset.
// Constant memory is cached, which should improve speed.
// The data is copied from constant memory to shared memory for the charset, this improves speed significantly.

__device__ __constant__ char deviceCharset[4096];
__device__ __constant__ __align__(16) unsigned char charsetLengths[16];
__device__ __constant__ unsigned char deviceSearchHashes[8192];
__device__ __constant__ unsigned char constantBitmap[8192]; // for lookups
__device__ __constant__ unsigned int hashTableShift;

unsigned char charsetLengthsHost[16] = {36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36};

// Include hash functions
#include "HashCounters.cu"
#include "HashCommon.cu"
#include "MD4.cu"
#include "MD5.cu"
#include "NTLM.cu"
#include "Multiforcer-common.c"
#include "SHA1.cu"

// Set up hash functions

// Yes.  I'm using global variables for this.  Byte me.
int hash_type;

// Define the hash types we support and number of hash types
char* Hash_Types[] = {"MD5", "FASTMD5", "MD4", "FASTMD4", "NTLM", "FASTNTLM", "SHA1", "FASTSHA1"};
char *Hash_Names[] = {"MD5", "MD5", "MD4", "MD4", "NTLM", "NTLM", "SHA1", "SHA1"};
#define NUMBER_OF_HASH_TYPES 8 

// Hash length in bytes
int Hash_Length[NUMBER_OF_HASH_TYPES] = {
	16,
	16,
	16,
	16,
	16,
	16,
	20,
	20
};

// Salt length in bytes (or zero for unsalted)
int Max_Salt_Length[NUMBER_OF_HASH_TYPES] = {
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0
};

// Launch_CUDA_Kernel function handles launching the CUDA kernel when given the right parameters.
typedef void (*Launch_CUDA_Kernel_Function)
	(int, uint32_t, int, unsigned char *, unsigned char *, struct start_positions *, uint32_t, uint32_t, uint32_t, unsigned char *, unsigned char *);
Launch_CUDA_Kernel_Function Launch_CUDA_Kernel[NUMBER_OF_HASH_TYPES] = {
	&Launch_CUDA_MD5_Kernel, 
	&Launch_CUDA_FASTMD5_Kernel, 
	&Launch_CUDA_MD4_Kernel, 
	&Launch_CUDA_FASTMD4_Kernel, 
	&Launch_CUDA_NTLM_Kernel, 
	&Launch_CUDA_FASTNTLM_Kernel, 
	&Launch_CUDA_SHA1_Kernel,
	&Launch_CUDA_FASTSHA1_Kernel};

// Gets the thread count to be used for a specific kernel
typedef int (*CUDA_Get_Thread_Count_Function)
	(int);
CUDA_Get_Thread_Count_Function CUDA_Get_Thread_Count[NUMBER_OF_HASH_TYPES] = {
	&CUDA_MD5_Get_Thread_Count, 
	&CUDA_MD5_Get_Thread_Count, 	
	&CUDA_MD4_Get_Thread_Count, 
	&CUDA_NTLM_Get_Thread_Count,
	NULL,
	NULL,
	NULL,
	NULL	
};

// Launch_CPU_Kernel function handles launching the CPU kernels for shorter lengths
typedef void (*Launch_CPU_Kernel_Function)
	(int, unsigned char *, unsigned char *, unsigned char *, unsigned char *, int, int);
Launch_CPU_Kernel_Function Launch_CPU_Kernel[NUMBER_OF_HASH_TYPES] = {
	&Launch_CPU_MD5_Kernel, 
	&Launch_CPU_MD5_Kernel, 
	&Launch_CPU_MD4_Kernel, 
	&Launch_CPU_MD4_Kernel, 
	&Launch_CPU_NTLM_Kernel,
	&Launch_CPU_NTLM_Kernel,
	&Launch_CPU_SHA1_Kernel,
	&Launch_CPU_SHA1_Kernel
};

// Takes path, hash length.  Loads hashes or errors.
// Returns sorted hash list.
typedef struct HOST_Hash_Data* (*Load_Hash_Function)
	(const char *, int);
Load_Hash_Function Load_Hashes[NUMBER_OF_HASH_TYPES] = {
	&Load_Unsalted_Hashes,
	&Load_Unsalted_Hashes,
	&Load_Unsalted_Hashes,
	&Load_Unsalted_Hashes,
	&Load_Unsalted_Hashes,
	&Load_Unsalted_Hashes,
	&Load_Unsalted_Hashes,
	&Load_Unsalted_Hashes	
};


/* printCudaDeviceInfo prints some basic specifications about the device. 
 * It returns the number of cores in the device, for detecting very low capability devices.
 */
int printCUDADeviceInfo(int device) {
  unsigned int free, total;
  CUresult res;
  CUdevice dev;
  CUcontext ctx;
  cudaDeviceProp deviceProp;

  cudaGetDeviceProperties(&deviceProp, device);

  cuDeviceGet(&dev,device);
  cuCtxCreate(&ctx, 0, dev);
  res = cuMemGetInfo(&free, &total);
  if(res != CUDA_SUCCESS)
      printf("!!!! cuMemGetInfo failed! (status = %x)", res);
  cuCtxDetach(ctx);

  printf("CUDA Device Information:\n");
  printf("Device %d: \"%s\"\n", device, deviceProp.name);
  printf("  Number of cores:                               %d\n", 8 * deviceProp.multiProcessorCount);
  printf("  Clock rate:                                    %.2f GHz\n", deviceProp.clockRate * 1e-6f);
  printf("  Memory (total):                                %lu MB\n", (total / (1024 * 1024)));
  printf("  Memory (free):                                 %lu MB\n\n", (free / (1024 * 1024)));
  return deviceProp.multiProcessorCount;
}


/* setStartPoints sets the initial positions for each thread.  p0 - p[len-1] are the positions in the character set that each thread tests.
 * By setting them here, we avoid doing 64-bit divides on the GPU, and this is done "for free" while the GPU works.
 * This should look very similar to a base conversion routine (it converts to base [charsetlength]).  
 */

void setStartPoints(start_positions *HOST_Start_Positions, uint64_t perThread, uint64_t numberOfThreads, 
				    uint64_t charsetLength, uint64_t start_point, int passLength) {

    uint64_t thread, thread_start;
	for (thread = 0; thread < numberOfThreads; thread++) {
	  // Where in the password space the thread starts: ThreadID * number per thread + current start offset
      thread_start = thread * perThread + start_point;  
	  // Start at the right point for the password length and fall through to the end   
      switch (passLength) {
	    case 16:
	      HOST_Start_Positions[thread].p15 = (unsigned char)(thread_start % charsetLength);
          thread_start /= charsetLength;
	    case 15:
	      HOST_Start_Positions[thread].p14 = (unsigned char)(thread_start % charsetLength);
          thread_start /= charsetLength;		  
	    case 14:
	      HOST_Start_Positions[thread].p13 = (unsigned char)(thread_start % charsetLength);
          thread_start /= charsetLength;		  
	    case 13:
	      HOST_Start_Positions[thread].p12 = (unsigned char)(thread_start % charsetLength);
          thread_start /= charsetLength;		  
	    case 12:
	      HOST_Start_Positions[thread].p11 = (unsigned char)(thread_start % charsetLength);
          thread_start /= charsetLength;		  
	    case 11:
	      HOST_Start_Positions[thread].p10 = (unsigned char)(thread_start % charsetLength);
          thread_start /= charsetLength;		  
	    case 10:
	      HOST_Start_Positions[thread].p9 = (unsigned char)(thread_start % charsetLength);
          thread_start /= charsetLength;		  
	    case 9:
	      HOST_Start_Positions[thread].p8 = (unsigned char)(thread_start % charsetLength);
          thread_start /= charsetLength;		  
	    case 8:
	      HOST_Start_Positions[thread].p7 = (unsigned char)(thread_start % charsetLength);
          thread_start /= charsetLength;		  
	    case 7:
	      HOST_Start_Positions[thread].p6 = (unsigned char)(thread_start % charsetLength);
          thread_start /= charsetLength;		  
	    case 6:
	      HOST_Start_Positions[thread].p5 = (unsigned char)(thread_start % charsetLength);
          thread_start /= charsetLength;		  
	    case 5:
	      HOST_Start_Positions[thread].p4 = (unsigned char)(thread_start % charsetLength);
          thread_start /= charsetLength;		  
	    case 4:
	      HOST_Start_Positions[thread].p3 = (unsigned char)(thread_start % charsetLength);
          thread_start /= charsetLength;		  
	    // Lengths 0-3 are handled by CPU, no case needed
        HOST_Start_Positions[thread].p2 = (unsigned char)(thread_start % charsetLength);
        thread_start /= charsetLength;
        HOST_Start_Positions[thread].p1 = (unsigned char)(thread_start % charsetLength);
        thread_start /= charsetLength;
        HOST_Start_Positions[thread].p0 = (unsigned char)(thread_start % charsetLength);
	  }
    }
}

// Deal with multiple bases - this will work for a single base too, but is substantially less efficient.

void setStartPointsMulti(start_positions *HOST_Start_Positions, uint64_t perThread, uint64_t numberOfThreads, 
				    unsigned char *charset_length_array, uint64_t start_point, int passLength) {

    uint64_t thread, thread_start;
    for (thread = 0; thread < numberOfThreads; thread++) {
	// Where in the password space the thread starts: ThreadID * number per thread + current start offset
        thread_start = thread * perThread + start_point;  
	// Start at the right point for the password length and fall through to the end   
      switch (passLength) {
	    case 16:
	      HOST_Start_Positions[thread].p15 = (UINT4)(thread_start % charset_length_array[15]);
          thread_start /= charset_length_array[15];
	    case 15:
	      HOST_Start_Positions[thread].p14 = (UINT4)(thread_start % charset_length_array[14]);
          thread_start /= charset_length_array[14];		  
	    case 14:
	      HOST_Start_Positions[thread].p13 = (UINT4)(thread_start % charset_length_array[13]);
          thread_start /= charset_length_array[13];		  
	    case 13:
	      HOST_Start_Positions[thread].p12 = (UINT4)(thread_start % charset_length_array[12]);
          thread_start /= charset_length_array[12];		  
	    case 12:
	      HOST_Start_Positions[thread].p11 = (UINT4)(thread_start % charset_length_array[11]);
          thread_start /= charset_length_array[11];		  
	    case 11:
	      HOST_Start_Positions[thread].p10 = (UINT4)(thread_start % charset_length_array[10]);
          thread_start /= charset_length_array[10];		  
	    case 10:
	      HOST_Start_Positions[thread].p9 = (UINT4)(thread_start % charset_length_array[9]);
          thread_start /= charset_length_array[9];		  
	    case 9:
	      HOST_Start_Positions[thread].p8 = (UINT4)(thread_start % charset_length_array[8]);
          thread_start /= charset_length_array[8];		  
	    case 8:
	      HOST_Start_Positions[thread].p7 = (UINT4)(thread_start % charset_length_array[7]);
          thread_start /= charset_length_array[7];		  
	    case 7:
	      HOST_Start_Positions[thread].p6 = (UINT4)(thread_start % charset_length_array[6]);
          thread_start /= charset_length_array[6];		  
	    case 6:
	      HOST_Start_Positions[thread].p5 = (UINT4)(thread_start % charset_length_array[5]);
          thread_start /= charset_length_array[5];		  
	    case 5:
	      HOST_Start_Positions[thread].p4 = (UINT4)(thread_start % charset_length_array[4]);
          thread_start /= charset_length_array[4];		  
	    case 4:
	      HOST_Start_Positions[thread].p3 = (UINT4)(thread_start % charset_length_array[3]);
          thread_start /= charset_length_array[3];		  
	    case 3:
          HOST_Start_Positions[thread].p2 = (UINT4)(thread_start % charset_length_array[2]);
          thread_start /= charset_length_array[2];
	    case 2:
          HOST_Start_Positions[thread].p1 = (UINT4)(thread_start % charset_length_array[1]);
          thread_start /= charset_length_array[1];
	    case 1:
          HOST_Start_Positions[thread].p0 = (UINT4)(thread_start % charset_length_array[0]);
	  } 
    }
}

/* outputFoundHashes takes the various arrays, looks for newly found passwords, prints them, and outputs them to the output file if needed.
 * It returns the number of passwords it found that were new.
 */
int outputFoundHashes(int numberOfPasswords, int passlength, unsigned char *HOST_Success, unsigned char *HOST_Success_Printed, unsigned char *HOST_SearchHashes, unsigned char *HOST_Passwords, FILE *outputfile) {
  int i, j;
  int passwordsFound = 0;
  
  for (i = 0; i < numberOfPasswords; i++) {
    if (HOST_Success[i] && !HOST_Success_Printed[i]) {
	  if (passwordsFound == 0) {
	    //clear line
		printf("                                                                               \r");
	  }
	  printf("%s : ", Hash_Names[hash_type]);
	  passwordsFound++;
      HOST_Success_Printed[i] = 1;
      for (j = 0; j < Hash_Length[hash_type]; j++) {
        printf("%02X", HOST_SearchHashes[i * Hash_Length[hash_type] + j]);
      }
      printf(": '");
	  for (j = 0; j < passlength; j++) {
	    printf("%c", HOST_Passwords[16*i+j]);
	  }
	  printf("'    0x");
	  for (j = 0; j < passlength; j++) {
	    printf("%02X", HOST_Passwords[16*i+j]);
	  }
	  printf("\n");
      /*
      if (outputfile) {
        fprintf(outputfile, "%s : ", Hash_Types[hash_type]);
        for (j = 0; j < 16; j++) {
          fprintf(outputfile, "%02X", HOST_SearchHashes[i * 16 + j]);
        }
        fprintf(outputfile, ": '");
	    for (j = 0; j < passlength; j++) {
	      fprintf(outputfile, "%c", HOST_Passwords[16*i+j]);
	    }
        fprintf(outputfile, "'    0x");
	    for (j = 0; j < passlength; j++) {
	      fprintf(outputfile, "%02X", HOST_Passwords[16*i+j]);
	    }
	    fprintf(outputfile, "\n");
	    fflush(outputfile);
      }
	  */
      if (outputfile) {
        fprintf(outputfile, "%s:", Hash_Names[hash_type]);
        for (j = 0; j < Hash_Length[hash_type]; j++) {
          fprintf(outputfile, "%02X", HOST_SearchHashes[i * Hash_Length[hash_type] + j]);
        }
        fprintf(outputfile, ":");
	    for (j = 0; j < passlength; j++) {
	      fprintf(outputfile, "%c", HOST_Passwords[16*i+j]);
	    }
	    fprintf(outputfile, "\n");
	    fflush(outputfile);
      }
	}
  }
  return passwordsFound; 
}



/*
 * Get the rate of the given kernel (in M/s) with the given parameters. 
 * This iterates through speeds until the target time is found, calculates the speed, and returns.
 */
uint32_t getKernelRate(int passlength, uint32_t charsetLength, int numberOfPasswords, int target_ms, uint32_t threads, 
					unsigned char *DEVICE_Passwords, unsigned char *DEVICE_Success) {
  uint64_t perThread;
  // Default of 8, for very low end GPUs if the first test fails
  uint32_t maxblocks = 8, blocks;
  unsigned int timer, per_step;
  float maxRate = 0.0, ref_time;
  int i;
  struct start_positions *DEVICE_Start_Positions, *HOST_Start_Positions;   
  unsigned char *HOST_Passwords, *HOST_Success;


  HOST_Success = (unsigned char *)malloc(numberOfPasswords*sizeof(unsigned char));
  HOST_Passwords = (unsigned char *)malloc(MAX_PASSWORD_LENGTH*numberOfPasswords*sizeof(unsigned char));

  printf("Please wait... finding optimum block count.\n");
  // Create & start a timer for the no-work time (we cannot go lower than this)
  cutCreateTimer(&timer);
  
  // Iterate through block counts in increments of 32.
  // If the kernel cannot execute in the specified time, quit.
  // Otherwise, find the max performance.
  for (blocks = 16; blocks <= 1024; blocks+=16) {
    // See main code for comments
	perThread = 1;
    for (i = 0; i < passlength; i++) {
      perThread *= (charsetLength);
    }
    perThread /= ((blocks * threads) - 1);
    // Allocate memory for device start positions
	HOST_Start_Positions = (struct start_positions *)malloc(blocks * threads * sizeof(struct start_positions));
    cudaMalloc((void **)&DEVICE_Start_Positions, blocks * threads * sizeof(struct start_positions));
    cudaMemset(DEVICE_Start_Positions, 0, blocks * threads * sizeof(struct start_positions));
	if (verbose) {
	  printf("Testing %d threads, %d blocks\n", (unsigned int)threads, (unsigned int)blocks);
	}
	// Null work test: See how long it takes to just launch the kernels.
	// If this is beyond our goal time, this is definitely too many threads.
	per_step = 0;
    cutResetTimer(timer);
	cutStartTimer(timer);
    cudaMemcpy(DEVICE_Start_Positions, HOST_Start_Positions, blocks * threads * sizeof(struct start_positions), cudaMemcpyHostToDevice);  
	//(*Launch_CUDA_Kernel[hash_type])(passlength, charsetLength, numberOfPasswords, DEVICE_Passwords, DEVICE_Success, DEVICE_Start_Positions, per_step, threads, blocks);
	cudaThreadSynchronize();
	cudaMemcpy(HOST_Success, DEVICE_Success, numberOfPasswords*sizeof(unsigned char), cudaMemcpyDeviceToHost);
    cudaMemcpy(HOST_Passwords, DEVICE_Passwords, MAX_PASSWORD_LENGTH*numberOfPasswords*sizeof(unsigned char), cudaMemcpyDeviceToHost);	
	cudaThreadSynchronize();
	cutStopTimer(timer);  
    ref_time = cutGetTimerValue(timer);
    if (verbose) {
      printf("Null work test done: Reference time %0.3f ms\n", ref_time);  
    }
	if (ref_time > target_ms) {
	  if (verbose) {
	    printf("Target execution time exceeded with 0 steps\n");
	  }
	  break;
	}
    // Try with per_step of 100 for reference
	per_step = 100;
    cutResetTimer(timer);
	cutStartTimer(timer);
    cudaMemcpy(DEVICE_Start_Positions, HOST_Start_Positions, blocks * threads * sizeof(struct start_positions), cudaMemcpyHostToDevice);  
	//(*Launch_CUDA_Kernel[hash_type])(passlength, charsetLength, numberOfPasswords, DEVICE_Passwords, DEVICE_Success, DEVICE_Start_Positions, per_step, threads, blocks);
	cudaThreadSynchronize();
	cudaMemcpy(HOST_Success, DEVICE_Success, numberOfPasswords*sizeof(unsigned char), cudaMemcpyDeviceToHost);
    cudaMemcpy(HOST_Passwords, DEVICE_Passwords, MAX_PASSWORD_LENGTH*numberOfPasswords*sizeof(unsigned char), cudaMemcpyDeviceToHost);	
	cudaThreadSynchronize();
	cutStopTimer(timer);  
    ref_time = cutGetTimerValue(timer);
    if (verbose) {
	  printf("100 step test done: Reference time %0.3f ms\n", ref_time);  
	}
	if (ref_time > (target_ms * 4.0)) {
	  if (verbose) {
	    printf("Target execution time exceeded by 4x on 100 steps, we are past the peak.\n");
	  }
	  break;
	}
    // Adjust to desired kernel ms
	per_step = (unsigned int) (per_step * ((float)target_ms / ref_time));
        if (verbose) {
          printf("Per step adjusted to %d\n", per_step);
        }
	cutResetTimer(timer);
	cutStartTimer(timer);
    cudaMemcpy(DEVICE_Start_Positions, HOST_Start_Positions, blocks * threads * sizeof(struct start_positions), cudaMemcpyHostToDevice);  
	//(*Launch_CUDA_Kernel[hash_type])(passlength, charsetLength, numberOfPasswords, DEVICE_Passwords, DEVICE_Success, DEVICE_Start_Positions, per_step, threads, blocks);
	cudaThreadSynchronize();
	cudaMemcpy(HOST_Success, DEVICE_Success, numberOfPasswords*sizeof(unsigned char), cudaMemcpyDeviceToHost);
    cudaMemcpy(HOST_Passwords, DEVICE_Passwords, MAX_PASSWORD_LENGTH*numberOfPasswords*sizeof(unsigned char), cudaMemcpyDeviceToHost);	
	cudaThreadSynchronize();
	cutStopTimer(timer);  
    ref_time = cutGetTimerValue(timer);
    if (verbose) {
      printf("%d step test done: Reference time %0.3f ms, rate %0.1fM/s\n", per_step, ref_time, (blocks * threads * per_step) / (ref_time * 1000.0));  
    }
	if (((blocks * threads * per_step) / (ref_time * 1000.0)) > maxRate) {
	  maxRate = (float) ((blocks * threads * per_step) / (ref_time * 1000.0));
	  maxblocks = blocks;
	}
	free(HOST_Start_Positions);
	cudaFree(DEVICE_Start_Positions);
	if (!verbose) {
	  printf(".");
	}
	fflush(stdout);
  }
  free(HOST_Success);
  free(HOST_Passwords);
  printf("\nPeak rate %0.1fM/s, block count %d\n", maxRate, maxblocks);
  return maxblocks;  
}


void launchComputeKernel(struct HOST_Run_Data *Run_Data, struct HOST_Hash_Data *Hash_Data) {

  // Passwords on the device and space on the host.
  unsigned char *DEVICE_Passwords, *DEVICE_Hashes;
  // Success flags for the host & GPU
  unsigned char *DEVICE_Success;
  // Timer related
  unsigned int timer, timer_step;
  float ref_time;
  // Counters
  int i;
  // These are all uint64_t to prevent underflow issues.
  // They must remain 64-bit or things break.
  //uint64_t blocks, threads;
  uint64_t perThread, start_point = 0, keySpace;
  uint64_t step_count = 0;
  uint32_t per_step;
  // Starting positions for the threads
  struct start_positions *DEVICE_Start_Positions, *HOST_Start_Positions;   
  
   // Thread count per GPU per occupancy calculator
// If the user has not specified threads, set this.
  if (!Run_Data->threads) {
    Run_Data->threads = (*CUDA_Get_Thread_Count[hash_type])(Run_Data->password_length);
  }
  // Default number per steps.  As this is updated quickly, this just needs to be in the ballpark.
  per_step = 100;
	
	 


  // If the password length is below 4, we deal with it on the CPU.
  if (Run_Data->password_length <= 0) {
    printf("Launching CPU kernel for password length %ld\n", Run_Data->password_length);
    (*Launch_CPU_Kernel[hash_type])(Run_Data->password_length, Hash_Data->hashes, 
    			Hash_Data->passwords, Hash_Data->success, Run_Data->charset, Run_Data->charset_length, Hash_Data->number_of_hashes);
    outputFoundHashes(Hash_Data->number_of_hashes, Run_Data->password_length, 
    		Hash_Data->success, Hash_Data->previous_success, Hash_Data->hashes, Hash_Data->passwords, Run_Data->outputfile); 
	return;
  }
  
  // Allocate memory for the GPU
  cudaMalloc((void **)&DEVICE_Hashes, Hash_Data->number_of_hashes * Hash_Data->hash_length * sizeof(unsigned char));
  cudaMemcpy(DEVICE_Hashes, Hash_Data->hashes, Hash_Data->number_of_hashes * Hash_Data->hash_length * sizeof(unsigned char), cudaMemcpyHostToDevice);  
  
  cudaMalloc((void **)&DEVICE_Success, Hash_Data->number_of_hashes*sizeof(unsigned char));
  cudaMemset(DEVICE_Success, 0, Hash_Data->number_of_hashes*sizeof(unsigned char));
  
  if (cudaMalloc((void **)&DEVICE_Passwords, MAX_PASSWORD_LENGTH*Hash_Data->number_of_hashes*sizeof(unsigned char)) == cudaErrorMemoryAllocation) {
    fprintf(stderr, "ERROR: CUDA memory allocation: Do you have enough free GPU memory?\n");
	exit(1);  
  }
  cudaMemset(DEVICE_Passwords, 0, MAX_PASSWORD_LENGTH*Hash_Data->number_of_hashes*sizeof(unsigned char));
  cudaError_t err = cudaGetLastError();
  if( cudaSuccess != err) 
    {
      fprintf(stderr, "Cuda error: %s.\n", cudaGetErrorString( err) );
      exit(EXIT_FAILURE);
    }

 // If the user has not specified blocks, tune.
  if (!Run_Data->blocks) {
    //blocks = getKernelRate(passlength, charsetLength, numberOfPasswords, target_ms, threads, DEVICE_Passwords, DEVICE_Success);
  }
  
  HOST_Start_Positions = (struct start_positions *)malloc(Run_Data->blocks * Run_Data->threads * sizeof(struct start_positions));
  
  // Check the last cudaMalloc for failure
  if (cudaMalloc((void **)&DEVICE_Start_Positions, Run_Data->blocks * Run_Data->threads * sizeof(struct start_positions)) == cudaErrorMemoryAllocation) {
    fprintf(stderr, "ERROR: CUDA memory allocation: Do you have enough free GPU memory?\n");
	exit(1);
  }
  cudaMemset(DEVICE_Start_Positions, 0, Run_Data->blocks * Run_Data->threads * sizeof(struct start_positions)); 

  


  // Calculate number of passwords per thread
  // pow() does not like uint64_t values.  
  perThread = 1;
  /*
  for (i = 0; i < Run_Data->password_length; i++) {
    perThread *= (Run_Data->charset_length);
  }
  */
  for (i = 0; i < Run_Data->password_length; i++) {
    if (Run_Data->charset_length_array[i] > 0) {
      perThread *= (Run_Data->charset_length_array[i]);
    }
  }
  keySpace = perThread;
  
  // To prevent integer truncation from causing missed passwords,
  // perThread is made slightly larger than needed.  This does not add
  // much work, but ensures 100% coverage of password space
  perThread /= Run_Data->blocks * Run_Data->threads;
  
  perThread += 1;

  // Create timer
  cutCreateTimer(&timer);
  cutCreateTimer(&timer_step);
  cutStartTimer(timer);
  
  if (verbose) {
    printf("Kernel launch parameters\n");
	printf("Blocks:      %d\n", Run_Data->blocks);
	printf("Threads:     %d\n", Run_Data->threads);
	printf("Passwords per thread: %ld\n\n", perThread);
  }
  
 
  printf("Launching kernel for password length %ld\n", Run_Data->password_length);
  
  while (start_point <= perThread) {
    step_count++;
	
    // Set the start_positions struct value appropriately for where we are
    /*
	setStartPoints(HOST_Start_Positions, perThread, (Run_Data->blocks * Run_Data->threads), 
    		Run_Data->charset_length, start_point, Run_Data->password_length);
    */
	setStartPointsMulti(HOST_Start_Positions, perThread, (Run_Data->blocks * Run_Data->threads), 
    		Run_Data->charset_length_array, start_point, Run_Data->password_length);
    // We sync here and wait for the GPU to finish.
    cudaThreadSynchronize();
    cutStopTimer(timer_step);  
	 
    // Wait for it to stabilize slightly before getting timer values, to handle large target_ms values
    if (step_count > 5) {
      ref_time = cutGetTimerValue(timer_step);
    } else {
      ref_time = 0;
    }
    cutResetTimer(timer_step);
	
    // Run this roughly every second, or every step if target_ms is >500
    if ((Run_Data->target_ms > 500) || (step_count % (1000 / Run_Data->target_ms) == 0)) { 
      // Copy device success & password lists to the host, look for new successes, and print them out "on the fly" (and, optionally, to the output file)
      cudaMemcpy(Hash_Data->success, DEVICE_Success, Hash_Data->number_of_hashes*sizeof(unsigned char), cudaMemcpyDeviceToHost);
      cudaMemcpy(Hash_Data->passwords, DEVICE_Passwords, MAX_PASSWORD_LENGTH*Hash_Data->number_of_hashes*sizeof(unsigned char), cudaMemcpyDeviceToHost);
	
      outputFoundHashes(Hash_Data->number_of_hashes, Run_Data->password_length, Hash_Data->success, 
      			Hash_Data->previous_success, Hash_Data->hashes, Hash_Data->passwords, Run_Data->outputfile);
	  if (verbose) {
        printf("Work: %d Done: %0.2f%% Time: %0.2f ms Stp/srch: %0.1fM/s %0.1fM/s  \r", 
      		per_step, (float)(100.0 * (float)start_point / (float)perThread), ref_time,
      		(float)((Run_Data->blocks * Run_Data->threads * (float)per_step) / (ref_time * 1000.0)), 
      		(float)(Hash_Data->number_of_hashes * (Run_Data->blocks * Run_Data->threads * (float)per_step) / (ref_time * 1000.0)));
        fflush(stdout);	  
	  } else {
	    printf("Done: %0.2f%%  Step rate: %0.1fM/s Search rate: %0.1fM/sec  \r", 
      		100.0 * (float)start_point / (float)perThread,
      		(Run_Data->blocks * Run_Data->threads * (float)per_step) / (ref_time * 1000.0), Hash_Data->number_of_hashes * (Run_Data->blocks * Run_Data->threads * (float)per_step) / (ref_time * 1000.0));
      fflush(stdout);
	  }  

      // If the current execution time is not correct, adjust.
	  
      if ((ref_time > 0) &&(ref_time < (Run_Data->target_ms * 0.9)) || (ref_time > (Run_Data->target_ms * 1.1))) {
        per_step = (uint32_t)((float)per_step * ((float)Run_Data->target_ms / ref_time));
		if (verbose) {printf("\nAdjusting passwords per step to %d\n", per_step);}
      }
    }
    cutStartTimer(timer_step);
    cudaMemcpy(DEVICE_Start_Positions, HOST_Start_Positions, Run_Data->blocks * Run_Data->threads * sizeof(struct start_positions), cudaMemcpyHostToDevice);  
    //cudaThreadSynchronize();
    (*Launch_CUDA_Kernel[hash_type])(Run_Data->password_length, Run_Data->charset_length, Hash_Data->number_of_hashes, DEVICE_Passwords, DEVICE_Success, DEVICE_Start_Positions, per_step, Run_Data->threads, Run_Data->blocks, DEVICE_Hashes, Hash_Data->global_bitmap_device);
    // Increment start point by however many we did
	start_point += per_step;
  }
  
  
  // Final copy of data over from the GPU for this run
  cudaThreadSynchronize();

  cudaMemcpy(Hash_Data->success, DEVICE_Success, Hash_Data->number_of_hashes*sizeof(unsigned char), cudaMemcpyDeviceToHost);
  cudaMemcpy(Hash_Data->passwords, DEVICE_Passwords, MAX_PASSWORD_LENGTH*Hash_Data->number_of_hashes*sizeof(unsigned char), cudaMemcpyDeviceToHost);
	
  outputFoundHashes(Hash_Data->number_of_hashes, Run_Data->password_length, Hash_Data->success, 
      			Hash_Data->previous_success, Hash_Data->hashes, Hash_Data->passwords, Run_Data->outputfile);


  cudaThreadSynchronize();
  cutStopTimer(timer);  
  ref_time = cutGetTimerValue(timer);
  printf("\n\n------------------------------------------\n\n");
  printf("Compute done: Reference time %0.1f seconds\n", ref_time / 1000.0);  
  printf("Stepping rate: %0.1fM %ss/s\n", keySpace / (ref_time * 1000.0), Hash_Names[hash_type]);
  printf("Search rate: %0.1fM %ss/s\n\n", (float) Hash_Data->number_of_hashes * (float) keySpace / (ref_time * 1000.0), Hash_Names[hash_type]);

  cudaFree(DEVICE_Success);
  cudaFree(DEVICE_Passwords);
  cudaFree(DEVICE_Start_Positions); 
  cudaFree(DEVICE_Hashes);
  free(HOST_Start_Positions);  
}


int main(int argc, char *argv[]) {
  
  FILE *charsetfile, *hashfile;
  int i;
  int deviceCount;
  
  struct HOST_Hash_Data *Hash_Data;
  struct HOST_Run_Data *Run_Data;
  
  Run_Data = (HOST_Run_Data *)malloc(sizeof(HOST_Run_Data));
  memset(Run_Data, 0, sizeof(HOST_Run_Data));
  // Allocate enough space for each position to have a unique charset
  Run_Data->charset = (unsigned char *)malloc(MAX_CHARSET * MAX_PASSWORD_LENGTH);
  memset(Run_Data->charset, 0, MAX_CHARSET * MAX_PASSWORD_LENGTH * sizeof(unsigned char));
  // Allocate space for the length of each unique charset
  Run_Data->charset_length_array = (unsigned char *)malloc(MAX_PASSWORD_LENGTH);
  memset(Run_Data->charset_length_array, 0, MAX_PASSWORD_LENGTH * sizeof(unsigned char));
  // Set default thread/block count, target MS count
  Run_Data->blocks = 128;
  Run_Data->threads = 64;
  Run_Data->target_ms = 50;
  
  printf("Cryptohaze.com CUDA Multiforcer (multiple hash brute forcer)\n");
  printf("by Bitweasil\n");
  
  printf("Version 0.7.1, length 0-14\n");
  printf("Currently supported hash types: ");
  for (i = 0; i < NUMBER_OF_HASH_TYPES; i++) {
    printf("%s ", Hash_Types[i]);
  }
  printf("\n");

  if (argc == 1)
  {
	  printf(
		  "\nREQUIRED Arguments:\n"
		  " -h <str> or --hashtype <str>\n"
		  "  Type of hash to crack\n\n"
		  " --min <int>\n"
		  "  Minimum password length\n\n"
		  " --max <int>\n"
		  "  Maximum password length\n\n"
		  " -f <file> or --hashfile <file>\n"
		  "  Hash file (one per line, ASCII-hex)\n\n\n"
		  "REQUIRED Arguments (pick one):\n"
		  " -c <file> or --charsetfile <file>\n"
		  "  Character set file (non-deliniated)\n\n"
		  " -u <file> or --charsetfilemulti <file>\n"
		  "  Character set file (newline between each line)\n\n\n"
		  "OPTIONAL Arguments:\n"
		  " -o <file> or --outputfile <file>\n"
		  "  Output file for results\n\n"
		  " -v --verbose\n"
		  "  Verbose output\n\n"
		  " -l <int> or --lookup <int>\n"
		  "  Big lookup table (useful for very long hash lists)\n"
		  "  This uses 2^<int> bits of video memory.\n"
		  "  (default 0 (no lookup table); min 23 (1 MiB); max 32 (512 MiB))\n\n"
		  " -d <int> or --device <int>\n"
		  "  CUDA device to use (default 0)\n\n"
		  " -m <int> or --ms <int>\n"
		  "  Target step time in ms (default 50)\n\n"
		  " -b <int> or --blocks <int>\n"
		  "  Number of thread blocks to run (default 128)\n"
		  "  Number of blocks should be number of multiprocessors because this program\n"
		  "  uses a lot of shared memory. You should not use a number that is not\n"
		  "  divisible by number of multiprocessors.\n"
		  "  number of multiprocessors = stream processors / 8\n"
		  "  This may change in future cards.\n\n"
		  " -t <int> or --threads <int>\n"
		  "  Number of threads per block (default 64)\n"
		  "  You shouldn't use a number that is not divisible by 16.\n\n"
		  " --maxthreads\n"
		  "  Use maximum number of threads possible. (Currently does nothing)\n\n"
		  " --autotune\n"
		  "  Autotune for maximum performance. (Currently does nothing)\n\n\n"
		  "For Best Performance:\n"
		  " Try out different -b, -t, and use the highest -l you can. Also try the FAST*\n"
		  " hash types if you don't have that many hashes. If you use FAST* don't use -l.\n"
		  " Best config I've found for a 9800 GTX+ with 50000 hashes is:\n"
		  " -b 16 -t 320 -l 31\n");
#ifdef _WIN32
	  // This is to prevent stupid questions in the forum that look like this.
	  // "WTF a window pops up and disappears right away?"
	  // "you need to run it in command prompt"
	  // ... twenty back and forth posts later...
	  // "what is cmd?"
	  // ":shock: start, run, type cmd"
	  printf("\n\nThis program must be ran from command prompt. Go to the start menu, then run, type \"cmd\" and hit ok. "
		  "Type in the full path for CUDA-Multiforcer.exe and then the arguments to run this program:\n"
		  "Ex:\n\"C:\\CUDA-Multiforcer\\CUDA-Multiforcer.exe\" --hashtype MD5 --min 1 --max 7 --hashfile \"C:\\md5s.txt\" --charsetfile \"C:\\CUDA-Multiforcer\\charsets\\charsetfile\"\n\n"
		  "Press enter to exit...");
	  getchar();
#endif
	  exit(1);
  }

  // Command line argument parsing with argtable
  struct arg_str *h = arg_str1("h", "hashtype", "{NTLM, MD4, MD5}", "type of hash to crack");
  struct arg_file *c = arg_file0("c", "charsetfile", "charsetfile", "charset file (non-deliniated)"); 
  struct arg_file *cm = arg_file0("u", "charsetfilemulti", "charsetfilemulti", "charset file (newline between each line)"); 
  struct arg_file *o = arg_file0("o", "outputfile", "outputfile", "output file for results");
  struct arg_file *f = arg_file1("f", "hashfile", "hashfile", "hash file (one per line, ASCII-hex)");
  struct arg_lit *v = arg_lit0("v", "verbose", "verbose output");
  struct arg_int *l = arg_int0("l", "lookup", "<n>", "BIG lookup table (useful for very long hash lists)");
  struct arg_int *min = arg_int1(NULL, "min", "<n>", "minimum password length");
  struct arg_int *max = arg_int1(NULL, "max", "<n>", "maximum password length");
  struct arg_int *d = arg_int0("d", "device", "<n>", "CUDA device to use");
  struct arg_int *m = arg_int0("m", "ms", "<n>", "target step time in ms");
  struct arg_int *b = arg_int0("b", "blocks", "<n>", "number of thread blocks to run");
  struct arg_int *t = arg_int0("t", "threads", "<n>", "number of threads per block");  
  struct arg_lit *mthreads = arg_lit0(NULL, "maxthreads", "use maximum number of threads possible");
  struct arg_lit *autotune = arg_lit0(NULL, "autotune", "autotune for maximum performance");
  struct arg_end *end = arg_end(20);
  void *argtable[] = {h,c,cm,o,f,v,l,min,max,d,m,b,t,mthreads,autotune,end};
  

  // Check for a supported CUDA device
  cudaGetDeviceCount(&deviceCount);
  if (deviceCount == 0) {
    fprintf(stderr, "ERROR: This program requires a CUDA-capable video card.\nNo cards found. Sorry. Exiting.\n");
    exit(1);
  }
 
 
  if (arg_nullcheck(argtable) != 0)
    fprintf(stderr, "ERROR: Insufficient memory\n"); 
  // Look for errors
  int nerrors = arg_parse(argc,argv,argtable);
  if (nerrors > 0) {
    // Print errors, exit.
    arg_print_errors(stdout,end,argv[0]);
    exit(1);
  }
 
  
  // General init section
  
  // Choose the hash type
  hash_type = -1;
  for (i = 0; i < NUMBER_OF_HASH_TYPES; i++) {
    if (strcmp(*h->sval, Hash_Types[i]) == 0) {
	  hash_type = i;
	  printf("Hash type: %s\n", Hash_Types[i]);
	}
  }
  if (hash_type == -1) {
    fprintf(stderr, "ERROR: Unknown hash type %s: Exiting.\n", *h->sval);
	exit(1);  
  }

  
  if (v->count) {
    verbose = 1;
    printf("Verbose mode enabled.\n");
  }
  if (l->count && *l->ival != 0)
  {
	  if (*l->ival < 23) {
		fprintf(stderr, "ERROR: Big lookup table must be greater than or equal to 23 (1 MiB).\n");
		exit(1);
	  }
	  if (*l->ival > 32) {
		fprintf(stderr, "ERROR: Big lookup table must be less than or equal to 32 (512 MiB).\n");
		exit(1);
	  } 
	  big_lookup_table = *l->ival;
  }
  if (*min->ival > *max->ival) {
    fprintf(stderr, "ERROR: Minimum password length (%d) greater than maximum password length (%d).\n", *min->ival, *max->ival);
    exit(1);
  }
  if (*min->ival < 0) {
    fprintf(stderr, "ERROR: Minimum password length must be greater than zero.\n");
    exit(1);
  }
  if (*max->ival > 16) {
    fprintf(stderr, "ERROR: Maximum password length must be less than or equal to 16.\n");
    exit(1);
  } 
  
  if (!b->count) {
    *b->ival = NULL;
  }
  if (!t->count) {
    *t->ival = NULL;
  }
  
  // Load files: Display errors and exit if not loaded.  
  charsetfile = NULL;
  // Check for single charset file
  if (c->count) {
    charsetfile = fopen(c->filename[0], "r");
    if (!charsetfile) {
      fprintf(stderr, "ERROR: Cannot open charset file %s\n", c->filename[0]);
      exit(1);
    }
  } else if(cm->count) {
    charsetfile = fopen(cm->filename[0], "r");
    if (!charsetfile) {
      fprintf(stderr, "ERROR: Cannot open multi charset file %s\n", cm->filename[0]);
      exit(1);
    }
  } else {
    fprintf(stderr, "ERROR: You must specify a charset or charset-multi file.\n");
	exit(1);
  }

  hashfile = fopen(f->filename[0], "r");
  if (!hashfile) {
    fprintf(stderr, "ERROR: Cannot open hash file %s\n", f->filename[0]);
    exit(1);
  }
  if (o->count) {
    // Append to any existing files.
    Run_Data->outputfile = fopen(o->filename[0], "a");
    if (!Run_Data->outputfile) {
      fprintf(stderr, "ERROR: Cannot open output file %s\n", o->filename[0]);
      exit(1);
    } 
  }

  // Set device if specified, print information
  if (d->count) {
    cudaSetDevice(*d->ival);
    printCUDADeviceInfo(*d->ival);
  } else {
    printCUDADeviceInfo(0);
  }

  if (m->count && (*m->ival > 0)) {
    Run_Data->target_ms = *m->ival;
	if (verbose) {
      printf("Target execution time: %d ms\n", Run_Data->target_ms);
    }
  }
  // Check for block or thread counts on the command line, set if used.
  if (b->count) {
    Run_Data->blocks = *b->ival;
  }
  if (t->count) {
    Run_Data->threads = *t->ival;
  }
  
  // If maxthreads is passed, set count to zero for auto-setting.
  // Otherwise, use 64 or the given amount.
  if (mthreads->count) {
    Run_Data->auto_threads = 0;
  }
  if (autotune->count) {
    Run_Data->auto_blocks = 0;
  }

  int hash_position = 0;
  int character;
  Run_Data->charset_length = 0;
  while (!feof(charsetfile) && (hash_position < MAX_PASSWORD_LENGTH)) {
    character = fgetc(charsetfile);
	if (character == EOF) {
	  break;
	} 
	Run_Data->charset[(MAX_CHARSET * hash_position) + Run_Data->charset_length_array[hash_position]] = character;
    // Deal with newline characters
    if ((Run_Data->charset[(MAX_CHARSET * hash_position) + Run_Data->charset_length_array[hash_position]] == 0xA) || 
		(Run_Data->charset[(MAX_CHARSET * hash_position) + Run_Data->charset_length_array[hash_position]] == 0xD)) {
      if (Run_Data->charset_length_array[hash_position] > Run_Data->charset_length) {
	    Run_Data->charset_length = Run_Data->charset_length_array[hash_position];
	  }
	  hash_position++;
    } else {
      Run_Data->charset_length_array[hash_position]++;
    }
  } 
  // Single charset loaded: Copy data to other positions
  if (c->count) {
    printf("Single charset loaded.\n");
    for (hash_position = 1; hash_position < MAX_PASSWORD_LENGTH; hash_position++) {
      Run_Data->charset_length_array[hash_position] = Run_Data->charset_length_array[0];
      memcpy(&Run_Data->charset[(MAX_CHARSET * hash_position)], &Run_Data->charset[0], MAX_CHARSET);
    }
  }
  
  if (verbose) {
    for (hash_position = 0; hash_position < MAX_PASSWORD_LENGTH; hash_position++) {
      printf("Character %d (%d characters): ", hash_position + 1, Run_Data->charset_length_array[hash_position]);
	  for (i = 0; i < Run_Data->charset_length_array[hash_position]; i++) {
	    printf("%c", Run_Data->charset[(MAX_CHARSET * hash_position) + i]);
	  }
	  printf("\n");
	}
  }

  printf("Loading & sorting hashes.  This may take a while.\n");
  Hash_Data = (*Load_Hashes[hash_type])(f->filename[0], Hash_Length[hash_type]);
  printf("Hashes loaded (%d hashes)\n", Hash_Data->number_of_hashes);


 
  /*================END OF INIT==========================*/
  
  // Put data on the card into shared memory
  cudaMemcpyToSymbol(deviceCharset, Run_Data->charset, 4096);
  cudaMemcpyToSymbol(constantBitmap, Hash_Data->shared_bitmap, 8192);
  cudaMemcpyToSymbol(charsetLengths, Run_Data->charset_length_array, 16);
  uint32_t tmp = 32 - big_lookup_table;
  cudaMemcpyToSymbol(hashTableShift, &tmp, 4);
  
  if ((Hash_Data->number_of_hashes * Hash_Length[hash_type]) <= 8192) {
    cudaMemcpyToSymbol(deviceSearchHashes, Hash_Data->hashes, 8192);
	printf("You may want to consider the FAST[hash] option if it exists.\n");
  }
  
  printf("Done with copy to constant\n");
  fflush(stdout);

  for (Run_Data->password_length = *min->ival; Run_Data->password_length <= *max->ival; Run_Data->password_length++) {
    if (Run_Data->charset_length_array[Run_Data->password_length - 1] == 0) {
      fprintf(stderr, "ERROR: Character set size at position %d is zero - did you mean to use -c instead of -u?\n", Run_Data->password_length);
      exit(1);
    }
    launchComputeKernel(Run_Data, Hash_Data);
  }

}
